#include "sapporo.h"

struct real3 {
  double x, y, z;
};

int main(int argc, char *argv[]) {
  int n = 1024;
  if (argc > 1) n = atoi(argv[1]);
  cerr << " n = " << n << endl;

  int nx = 1;
  int n1 = 0;
  int n2 = n;

  double (*pos)[3] = new double[n][3];
  double (*vel)[3] = new double[n][3];
  double (*acc)[3] = new double[n][3];
  double (*jrk)[3] = new double[n][3];
  int    *nnb  = new int[n];
  int    *nngb  = new int[n];
  int    (*ngb_list)[256] = new int[n][256];
  double *pot  = new double[n];
  double *h2  = new double[n];
  double *mass = new double[n];
  int    *id   = new int[n];

  double tm = 0;
  for (int i = 0; i < n; i++) {
    pos[i][0] = drand48();
    pos[i][1] = drand48();
    pos[i][2] = drand48();

    vel[i][0] = drand48();
    vel[i][1] = drand48();
    vel[i][2] = drand48();
    
    h2[i] = 1.0 * pow(8.0/n, 1.0/3);
    h2[i] = h2[i]*h2[i];

    mass[i] = 100.0/n * drand48();
    tm += mass[i];
    id[i] = i;
  }

  for (int i = 0; i < n; i++) {
    mass[i] *= 1.0/tm;
  }
  
  sapporo grav;
  
  int cluster_id;
  grav.open(cluster_id);
  
   int ipmax = grav.get_n_pipes();
  fprintf(stderr, "ipmax= %d\n", ipmax);
//   double *h2 = new double[ipmax];
  double *i_nene = new double[ipmax];

  double null3[3] = {0,0,0};
  for (int i = 0; i < n; i++) {
    grav.set_j_particle(cluster_id, 
			i, id[i],
			0, 0,
			mass[i], null3, null3, null3,
			vel[i], pos[i]);
  }
  double ti = 0.1;
  grav.set_ti(cluster_id, ti);
    

  double eps2 = 0.0;
  for (int kk = 0; kk < 2; kk++) {
    fprintf(stderr, "Computing forces on GPU\n");
    double t1 = get_time();
    int n_evaluate = 0;
    n1 = 0;
    n2 = n;
    n_evaluate += n2 - n1;

    int i = n1; 
    while (i < n2) {
//     for (int i = n1; i < n2; i += ipmax) {
//      ipmax = min((int)(drand48()*grav.get_n_pipes() + 10), int(0.5*grav.get_n_pipes()));
      int npart = min(n2 - i, ipmax);
      if (kk == 0) {
	for (int ik = i; ik < i + npart; ik++) {
	  double dt = ti;
	  pos[ik][0] += vel[ik][0] * dt;
	  pos[ik][1] += vel[ik][1] * dt;
	  pos[ik][2] += vel[ik][2] * dt;
	}
      }
      grav.calc_firsthalf(cluster_id,
			  n, npart,
			  id + i,
			  pos+i, vel+i,
			  acc+i, jrk+i, pot+i, eps2, h2+i);

#ifdef NGB
      grav.calc_lasthalf2(cluster_id,
			  n, npart,
			  id+i, pos+i, vel+i,
			  eps2, h2+i,
			  acc+i, jrk+i, pot+i, nnb+i);
      grav.read_ngb_list(cluster_id);
      for (int i1 = i; i1 < i + npart; i1++) {
	if (i1 == 1034) {
	  cerr << "i=11034\n";
	}
	grav.get_ngb_list(cluster_id,
			  i1 - i,
			  n,
			  nngb[i1],
			  ngb_list[i1]);
      }
#else
      grav.calc_lasthalf(cluster_id,
			  n, npart,
			  id+i, pos+i, vel+i,
			  eps2, h2+i,
			  acc+i, jrk+i, pot+i);
#endif
      i += npart;
    }
    double dt = get_time() - t1;
    fprintf(stderr, " Completed in %lf seconds [ni_total= %d  %lg GFLOP/s] \n", dt, n_evaluate, 60.0e-9*n_evaluate*n/dt);
  }
  grav.close(cluster_id);
    
  double (*accx)[3] = new double[n][3];
  double (*jrkx)[3] = new double[n][3];
  double *potx = new double[n];
  int    *nnbx  = new int[n];
  int    *nngbx = new int[n];

  fprintf(stderr, "Computing forces on CPU\n");
  double t1 = get_time();

  int *ngb_listx = new int[n];
  for (int i = nx*n1; i < nx*n2; i++) {
    if (i%100 == 0) fprintf(stderr, "i= %d [%d]\n", i, nx);
    double posi[3] = {pos[i][0], pos[i][1], pos[i][2]};
    double veli[3] = {vel[i][0], vel[i][1], vel[i][2]};

    double acct[4] = {0,0,0,0};
    double jrkt[3] = {0,0,0};
    double ds2_min = 1.0e10;
    int id_min = 0;
    int nngb_i = 0;

    for (int j = 0; j < n; j++) {
      double dr[3] = {posi[0] - pos[j][0],
		      posi[1] - pos[j][1],
		      posi[2] - pos[j][2]};

      double dv[3] = {veli[0] - vel[j][0],
		      veli[1] - vel[j][1],
		      veli[2] - vel[j][2]};
      double ds2 = dr[0]*dr[0] + dr[1]*dr[1] + dr[2]*dr[2];
      
      if (ds2 <= h2[i]) 
	ngb_listx[nngb_i++] = id[j];


      if (ds2 == 0) continue;

      if (ds2 < ds2_min) {
	ds2_min = ds2;
	id_min = id[j];
      }

      double s = sqrt(ds2 + eps2);
      double c = 1.0/s/s/s;
      
      acct[0] -= mass[j] * c * dr[0];
      acct[1] -= mass[j] * c * dr[1];
      acct[2] -= mass[j] * c * dr[2];
      acct[3] -= mass[j] / s;
      
      // jerk
      double s5 = c/s/s;
      double rv = 3.0 * s5 * (dr[0]*dv[0] + dr[1]*dv[1] + dr[2]*dv[2]);

      jrkt[0] -= mass[j] * (c * dv[0] - rv * dr[0]);
      jrkt[1] -= mass[j] * (c * dv[1] - rv * dr[1]);
      jrkt[2] -= mass[j] * (c * dv[2] - rv * dr[2]);
    }
    accx[i][0] = acct[0];
    accx[i][1] = acct[1];
    accx[i][2] = acct[2];
    potx[i]    = acct[3];
    
    jrkx[i][0] = jrkt[0];
    jrkx[i][1] = jrkt[1];
    jrkx[i][2] = jrkt[2];
    
    nnbx[i] = id_min;
    nngbx[i] = nngb_i;
    fprintf(stdout, "%d %d\n", nngb[i], nngbx[i]);
//     for (int k = 0; k < nngb_i; k++) {
//       fprintf(stdout,"%d %d \n", ngb_list[i][k], ngb_listx[k]);
//     }
//     fprintf(stdout, "*****************************\n");
  }
  cerr << " done in " << get_time() - t1 << " sec \n";

  
  double force[3]  = {0,0,0};
  double  torque[3] = {0,0,0};
  for (int i = n1; i < n2; i++) {
	    
    force[0] += mass[i] * acc[i][0];
    force[1] += mass[i] * acc[i][1];
    force[2] += mass[i] * acc[i][2];
    
    torque[0] += mass[i] * (acc[i][1]*pos[i][2] - acc[i][2]*pos[i][1]);
    torque[1] += mass[i] * (acc[i][2]*pos[i][0] - acc[i][0]*pos[i][2]);
    torque[2] += mass[i] * (acc[i][0]*pos[i][1] - acc[i][1]*pos[i][0]);
    
    if (nx == 0) continue;
    real3 da = (real3){acc[i][0] - accx[i][0],
		       acc[i][1] - accx[i][1],
		       acc[i][2] - accx[i][2]};

    real3 dj = (real3){jrk[i][0] - jrkx[i][0],
		       jrk[i][1] - jrkx[i][1],
		       jrk[i][2] - jrkx[i][2]};
    
    double dpot = pot[i] - potx[i];

    fprintf(stdout, "%lg %lg %lg %d %d %d %d\n",
	    sqrt((da.x*da.x + da.y*da.y+da.z*da.z)/(accx[i][0]*accx[i][0] + accx[i][1]*accx[i][1]+ accx[i][2]*accx[i][2])),
	    sqrt((dj.x*dj.x + dj.y*dj.y+dj.z*dj.z)/(jrkx[i][0]*jrkx[i][0] + jrkx[i][1]*jrkx[i][1]+ jrkx[i][2]*jrkx[i][2])),
	    fabs(dpot)/potx[i], nnb[i], nnbx[i], nngb[i], nngbx[i]);

//     fprintf(stdout, "%g %g %g  %g %g %g  %g\n",
// 	    acc[i][0], acc[i][1], acc[i][2],
// 	    jrk[i][0], jrk[i][1], jrk[i][2], pot[i]);
//     double a = pot[i];
//     double b = potx[i];
//     fprintf(stdout, " %g %g %g \n",
// 	    a, b, (a-b)/b);
//     fprintf(stdout, "%g %g %g  %g %g %g  %g\n",
// 	    accx[i][0], accx[i][1], accx[i][2],
// 	    jrkx[i][0], jrkx[i][1], jrkx[i][2], potx[i]);

  }

  fprintf(stderr, "total force  = [%lg %lg %lg] \n",
	  force[0], force[1], force[2]);
  fprintf(stderr, "total torque = [%lg %lg %lg] \n",
	  torque[0], torque[1], torque[2]);

 
  cerr << "done!\n";
}
